//
//  TSCommonTool.swift
//  TSLiveWallpaper
//
//  Created by 100Years on 2024/12/20.
//

import Kingfisher
import Foundation
import AVFoundation
open class TSCommonTool {
    /// 下载图片，优先从缓存中取，支持设置缓存策略
    /// - Parameters:
    ///   - url: 图片的 URL 地址
    ///   - completion: 完成回调，返回图片或错误
    public static func fetchImageWithCache(
        from urlString: String,
        completion: @escaping (UIImage?, Error?) -> Void
    ) {
        
        guard let url = URL(string: urlString) else{
            completion(nil, NSError(domain: "url null", code: 0))
            return
        }

        // 配置 Kingfisher 的缓存策略
        let options: KingfisherOptionsInfo = [
            .cacheOriginalImage,
            .memoryCacheExpiration(.days(7)),  // 内存缓存时间
            .diskCacheExpiration(.days(30)),  // 磁盘缓存时间
            .cacheSerializer(DefaultCacheSerializer.default)
        ]
        
        // 使用 Kingfisher 检查缓存是否存在
        let cache = ImageCache.default
        let cacheKey = url.absoluteString
        
        cache.retrieveImage(forKey: cacheKey, options: nil) { result in
            switch result {
            case .success(let value):
                if let image = value.image {
                    // 从缓存中取到图片
                    completion(image, nil)
                } else {
                    // 缓存中不存在图片，进行下载
                    downloadImage(from: url, options: options, completion: completion)
                }
            case .failure(let error):
                // 缓存检查失败，直接下载
                print("缓存检查失败: \(error.localizedDescription)")
                downloadImage(from: url, options: options, completion: completion)
            }
        }
    }
    
    
    /// 获取已缓存的磁盘图片文件路径
    /// - Parameters:
    ///   - urlString: 图片的 URL 地址
    ///   - completion: 完成回调，返回图片的磁盘路径或错误
    public static func fetchImagePathWithCache(
        from urlString: String,
        completion: @escaping (String?, Error?) -> Void
    ) {
        guard let url = URL(string: urlString) else {
            completion(nil, NSError(domain: "InvalidURL", code: 0, userInfo: [NSLocalizedDescriptionKey: "无效的 URL"]))
            return
        }
        
        let cache = ImageCache.default
        let cacheKey = url.absoluteString
        if cache.isCached(forKey: cacheKey) {
            completion(cache.cachePath(forKey: cacheKey), nil)
        }else {
          completion(nil, NSError(domain: "CachePathError", code: 1, userInfo: [NSLocalizedDescriptionKey: "无法获取缓存路径"]))
        }

    }
    

    /// 下载图片并缓存
    /// - Parameters:
    ///   - url: 图片的 URL 地址
    ///   - options: Kingfisher 下载和缓存的选项
    ///   - completion: 完成回调，返回图片或错误
    public static func downloadImage(
        from url: URL,
        options: KingfisherOptionsInfo,
        completion: @escaping (UIImage?, Error?) -> Void
    ) {
        KingfisherManager.shared.retrieveImage(with: url, options: options,progressBlock: { receivedSize, totalSize in
            debugPrint(receivedSize)
        }) { result in
            switch result {
            case .success(let value):
                completion(value.image, nil)
            case .failure(let error):
                completion(nil, error)
            }
        }
    }
    
    
    
    /// 下载并缓存文件，依据 URL 的后缀名动态设置文件名
    /// - Parameters:
    ///   - url: 文件的 URL 地址
    ///   - completion: 完成回调，返回本地缓存路径或错误
    public static func downloadAndCacheFile(from urlString: String, fileEx:String? = nil, cacheDirectory:String = "cacheAll",completion: @escaping (String?, Error?) -> Void) {
        
        guard let url = URL(string: urlString) else{
            completion(nil, NSError(domain: "url null", code: 0))
            return
        }
        
        
        if !urlString.contains("http") && urlString.contains("/"){
            completion(urlString.fillCachePath, nil)
            return
        }
        
        let fileManager = FileManager.default
        
        // 获取缓存目录下的 `cacheVideo` 文件夹路径
        let cachesDirectory = fileManager.urls(for: .cachesDirectory, in: .userDomainMask).first!
        let cacheVideoDirectory = cachesDirectory.appendingPathComponent(cacheDirectory)
        
        // 创建 `cacheVideo` 文件夹（如果不存在）
        if !fileManager.fileExists(atPath: cacheVideoDirectory.path) {
            do {
                try fileManager.createDirectory(at: cacheVideoDirectory, withIntermediateDirectories: true, attributes: nil)
            } catch {
                completion(nil, error)
                return
            }
        }
        
        var fileName = url.path.md5
        
        // 使用 URL 的 MD5 哈希值作为缓存文件名，附加 URL 的后缀名
        var fileExtension = fileEx
        fileExtension = fileExtension ?? (url.pathExtension.isEmpty ? "" : url.pathExtension)
        if let fileExtension = fileExtension,fileExtension.count > 0 {
            fileName = url.path.md5 + ".\(fileExtension)"
        }

        let cachedFileURL = cacheVideoDirectory.appendingPathComponent(fileName)
        
        //检查文件是否已存在于缓存中
//        if fileManager.fileExists(atPath: cachedFileURL.path) {
//            print("文件已存在于缓存中: \(cachedFileURL)")
//            completion(cachedFileURL.path, nil)
//            return
//        }
        
        // 下载文件
        let task = URLSession.shared.downloadTask(with: url) { tempFileURL, response, error in
            if let error = error {
                DispatchQueue.main.async {
                    completion(nil, error)
                }
                return
            }
            
            guard let tempFileURL = tempFileURL else {
                DispatchQueue.main.async {
                    completion(nil, NSError(domain: "DownloadError", code: -1, userInfo: [NSLocalizedDescriptionKey: "临时文件路径不存在"]))
                }
                return
            }
            
            do {
                if fileManager.fileExists(atPath: cachedFileURL.path) {
                    try fileManager.removeItem(atPath:cachedFileURL.path)
                    dePrint("下载成功,移除已有的旧文件: \(cachedFileURL)")
                }
                try fileManager.moveItem(at: tempFileURL, to: cachedFileURL)
                dePrint("文件下载并缓存成功: \(cachedFileURL)")
                DispatchQueue.main.async {
                    completion(cachedFileURL.path, nil)
                }
            } catch {
                dePrint("文件下载成功,但失败:\(error)")
                DispatchQueue.main.async {
                    completion(nil, error)
                }
            }
        }
        
        task.resume()
    }
}

public let kMainQueue = DispatchQueue.main
/// 主线程延迟执行回调
/// - Parameters:
///   - delay: 延迟时间（秒）
///   - completion: 延迟后的回调
public func kDelayOnMainThread(_ delay: TimeInterval, completion: @escaping () -> Void) {
    DispatchQueue.main.asyncAfter(deadline: .now() + delay) {
        completion()
    }
}

public func kDelayMainShort(completion: @escaping () -> Void) {
    DispatchQueue.main.asyncAfter(deadline: .now() + 0.1) {
        completion()
    }
}

/// 在主线程上执行代码
/// - Parameter block: 要执行的代码块
public func kExecuteOnMainThread(_ block: @escaping () -> Void) {
    if Thread.isMainThread {
        block()
    } else {
        DispatchQueue.main.async {
            block()
        }
    }
}


/// 子 size 根据容器 size,返回自己的最大 size
/// - Parameters:
///   - containerSize: 容器 Size
///   - subviewSize: 子 Size
///   - contentInset: contentInset
///   - contentMode: contentMode
/// - Returns: 最大 size
public func kCalculateScaledSize(containerSize: CGSize, subviewSize: CGSize, contentInset: UIEdgeInsets = .zero ,contentMode:UIView.ContentMode) -> CGSize {
    // 计算去除边距后的容器可用大小
    let availableWidth = containerSize.width - contentInset.left - contentInset.right
    let availableHeight = containerSize.height - contentInset.top - contentInset.bottom
    
    switch contentMode {
    case .scaleToFill:
        // 直接填充容器，忽略子视图比例
        return CGSize(width: availableWidth, height: availableHeight)
    case .scaleAspectFit:
        // 按比例适配容器，确保子视图完全显示在容器内
        let widthRatio = availableWidth / subviewSize.width
        let heightRatio = availableHeight / subviewSize.height
        let scale = min(widthRatio, heightRatio)
        return CGSize(width: subviewSize.width * scale, height: subviewSize.height * scale)
    case .scaleAspectFill:
        // 按比例填充容器，可能会裁剪子视图
        let widthRatio = availableWidth / subviewSize.width
        let heightRatio = availableHeight / subviewSize.height
        let scale = max(widthRatio, heightRatio)
        return CGSize(width: subviewSize.width * scale, height: subviewSize.height * scale)
    default:
        // 默认返回子视图原始大小
        return subviewSize
    }
}
